# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice-module for Tryton from m-ds.de.
# The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.model import fields, Index
from trytond.pool import PoolMeta, Pool
from trytond.pyson import Eval, Bool, And, Or
from trytond.modules.product import price_digits, round_price
from trytond.transaction import Transaction
from sql.aggregate import Sum
from sql.conditionals import Coalesce, Case
from sql.functions import Round
from decimal import Decimal
from .tax import GrossTaxableMixin

# TODO: rewrite Invoice.search_untaxed_amount() to avoid rounding errors
#       if search for gross-invoices

sel_modegross = [
    ('net', 'Net invoice'),
    ('gross', 'Gross invoice'),
    ]


class Invoice(GrossTaxableMixin, metaclass=PoolMeta):
    __name__ = 'account.invoice'

    modegross = fields.Selection(
        string='Invoice type', selection=sel_modegross, required=True,
        help='Calculate the invoice gross or net.',
        states={'readonly': Bool(Eval('lines'))}, depends=['lines'])

    @classmethod
    def __setup__(cls):
        super(Invoice, cls).__setup__()
        t = cls.__table__()
        cls._sql_indexes.update({
                Index(t, (t.modegross, Index.Equality())),
            })

    @classmethod
    def default_modegross(cls):
        """ get setting from config
        """
        Configuration = Pool().get('account.configuration')

        cfg1 = Configuration.get_singleton()
        return getattr(cfg1, 'modegross', 'net')

    @classmethod
    def get_amount(cls, invoices, names):
        """ fixup total_amount + untaxed_amount for gross-invoices
        """
        pool = Pool()
        Inv1 = pool.get('account.invoice')

        res1 = super(Invoice, cls).get_amount(
            invoices,
            ['total_amount', 'untaxed_amount', 'tax_amount'])

        inv_lst = Inv1.search([
                ('id', 'in', list(res1['total_amount'].keys())),
                ('modegross', '=', 'gross'),
                ('move', '=', None)])

        for invoice in inv_lst:
            res1['total_amount'][invoice.id] = sum(
                    (line.amount for line in invoice.lines
                        if line.type == 'line'), Decimal(0))
            res1['untaxed_amount'][invoice.id] = \
                res1['total_amount'][invoice.id] - \
                res1['tax_amount'][invoice.id]
        for key in list(res1.keys()):
            if key not in names:
                del res1[key]
        return res1

    @fields.depends(
        'lines', 'taxes', 'currency',
        'accounting_date', 'invoice_date', 'modegross',
        methods=['_get_taxes'])
    def _on_change_lines_taxes(self):
        """ calculate taxes, switch net/gross calculation
        """
        pool = Pool()
        InvoiceTax = pool.get('account.invoice.tax')

        modegross = getattr(self, 'modegross', 'net')
        if modegross == 'net':
            super(Invoice, self)._on_change_lines_taxes()
            return
        elif modegross != 'gross':
            raise ValueError('invalid "modegross"')

        # calculate in gross-mode
        self.untaxed_amount = Decimal('0.0')
        self.tax_amount = Decimal('0.0')
        self.total_amount = Decimal('0.0')
        computed_taxes = {}

        if self.lines:
            for line in self.lines:
                self.total_amount += getattr(line, 'amount', None) or 0
            computed_taxes = self._get_taxes_gross()

        def is_zero(amount):
            if self.currency:
                return self.currency.is_zero(amount)
            else:
                return amount == Decimal('0.0')

        tax_keys = []
        taxes = list(self.taxes or [])
        for tax in (self.taxes or []):
            if tax.manual:
                self.tax_amount += tax.amount or Decimal('0.0')
                continue
            key = tax._key
            if (key not in computed_taxes) or (key in tax_keys):
                taxes.remove(tax)
                continue
            tax_keys.append(key)
            if not is_zero(
                    computed_taxes[key]['base'] -
                    (tax.base or Decimal('0.0'))):
                self.tax_amount += computed_taxes[key]['amount']
                tax.amount = computed_taxes[key]['amount']
                tax.base = computed_taxes[key]['base']
            else:
                self.tax_amount += tax.amount or Decimal('0.0')
        for key in computed_taxes:
            if key not in tax_keys:
                self.tax_amount += computed_taxes[key]['amount']
                value = InvoiceTax.default_get(
                    list(InvoiceTax._fields.keys()), with_rec_name=False)
                value.update(computed_taxes[key])
                invoice_tax = InvoiceTax(**value)
                if invoice_tax.tax:
                    invoice_tax.sequence = invoice_tax.tax.sequence
                taxes.append(invoice_tax)
        self.taxes = taxes
        if self.currency:
            self.total_amount = self.currency.round(self.total_amount)
            self.tax_amount = self.currency.round(self.tax_amount)
        self.untaxed_amount = self.total_amount - self.tax_amount
        if self.currency:
            self.untaxed_amount = self.currency.round(self.untaxed_amount)

    @classmethod
    def search_total_amount(cls, name, clause):
        """ rewrite function to use 'modegross'
        """
        pool = Pool()
        Rule = pool.get('ir.rule')
        Line = pool.get('account.invoice.line')
        Tax = pool.get('account.invoice.tax')
        Invoice = pool.get('account.invoice')
        Currency = pool.get('currency.currency')
        type_name = cls.total_amount._field.sql_type().base
        line = Line.__table__()
        invoice = Invoice.__table__()
        currency = Currency.__table__()
        tax = Tax.__table__()

        _, operator, value = clause
        invoice_query = Rule.query_get('account.invoice')
        Operator = fields.SQL_OPERATORS[operator]

        union = (line.join(
                invoice,
                condition=(invoice.id == line.invoice)
            ).join(
                currency,
                condition=(currency.id == invoice.currency)
            ).select(
                line.invoice.as_('invoice'),
                Coalesce(
                    Sum(
                        Round(
                            (line.quantity * Case(
                                (invoice.modegross == 'gross',
                                    line.unit_gross_price),
                                else_=line.unit_price)).cast(type_name),
                            currency.digits)
                    ), 0).as_('total_amount'),
                where=line.invoice.in_(invoice_query),
                group_by=line.invoice
            ) | tax.join(
                invoice,
                condition=(tax.invoice == invoice.id),
            ).select(
                tax.invoice.as_('invoice'),
                Coalesce(
                    Sum(
                        Case(
                            (invoice.modegross == 'gross', Decimal('0.0')),
                            else_=tax.amount)
                    ), 0).as_('total_amount'),
                where=tax.invoice.in_(invoice_query),
                group_by=tax.invoice))
        query = union.select(
            union.invoice,
            group_by=union.invoice,
            having=Operator(
                Sum(union.total_amount).cast(type_name),
                value)
            )
        return [('id', 'in', query)]

# end Invoice


class InvoiceLine(GrossTaxableMixin, metaclass=PoolMeta):
    __name__ = 'account.invoice.line'

    modegross = fields.Function(
        fields.Char(
            string='Invoice type', readonly=True,
            states={'invisible': True}),
        'on_change_with_modegross')
    unit_gross_price = fields.Numeric(
        string='Gross Price', digits=price_digits,
        states={
            'invisible': Or(
                    Eval('type') != 'line',
                    Eval('modegross', '') != 'gross'),
            'required': And(
                    Eval('type') == 'line',
                    Eval('modegross', '') == 'gross'),
            'readonly': Eval('invoice_state') != 'draft'},
        depends=['type', 'modegross', 'invoice_state'])

    @classmethod
    def __setup__(cls):
        super(InvoiceLine, cls).__setup__()
        cls.unit_price.states.update({
            'required': And(
                Eval('type', '') == 'line',
                Eval('modegross', '') == 'net')})
        cls.unit_price.depends.add('modegross')

    def get_move_lines(self):
        """ move.lines for invoice.line
        """
        pool = Pool()
        Currency = pool.get('currency.currency')

        modegross = getattr(self, 'modegross', 'net')
        if modegross == 'net':
            return super(InvoiceLine, self).get_move_lines()
        elif modegross != 'gross':
            raise ValueError('invalid "modegross"')

        if self.type != 'line':
            return []

        # call .get_move_lines() from parent-class to not break the
        # inheritance-chain
        line, = super(InvoiceLine, self).get_move_lines()

        # in gross-mode: we cannot use 'net_price * quantity' -
        # because this will cause rounding errors,
        # 'base' is reverse calculated from gross_price and
        # per invoice-line rounding-corrected
        l1 = list(self._get_taxes().keys())
        if len(l1) == 0:
            # no tax: amount = amount
            amount = self.amount
        else:
            amount = l1[0]['base']

        if self.invoice.currency != self.invoice.company.currency:
            with Transaction().set_context(date=self.invoice.currency_date):
                amount = Currency.compute(
                    self.invoice.currency,
                    amount,
                    self.invoice.company.currency)
            line.amount_second_currency = amount
            line.second_currency = self.invoice.currency
        else:
            line.amount_second_currency = None
            line.second_currency = None
        if amount >= 0:
            if self.invoice.type == 'out':
                line.debit, line.credit = 0, amount
            else:
                line.debit, line.credit = amount, 0
        else:
            if self.invoice.type == 'out':
                line.debit, line.credit = -amount, 0
            else:
                line.debit, line.credit = 0, -amount
        if line.amount_second_currency:
            line.amount_second_currency = (
                line.amount_second_currency.copy_sign(
                    line.debit - line.credit))
        line.account = self.account
        if self.account.party_required:
            line.party = self.invoice.party
        line.origin = self
        line.tax_lines = self._compute_taxes()

        # update values from net-calculation to gross-mode
        if hasattr(line, 'analytic_lines') and hasattr(
                self, 'analytic_accounts'):
            analytic_lines = []
            line_date = (
                self.invoice.accounting_date or self.invoice.invoice_date)
            for entry in self.analytic_accounts:
                analytic_lines.extend(
                    entry.get_analytic_lines(line, line_date))
            line.analytic_lines = analytic_lines
        return [line]

    @property
    def taxable_lines(self):
        """ get taxable line, switch gross/net calculation
        """
        if getattr(self, 'invoice', None) and \
                getattr(self.invoice, 'modegross', None):
            modegross = self.invoice.modegross
        else:
            modegross = getattr(self, 'modegross', 'net')

        if modegross == 'net':
            return super(InvoiceLine, self).taxable_lines
        elif modegross == 'gross':
            context = Transaction().context
            if (getattr(self, 'invoice', None)
                    and getattr(self.invoice, 'type', None)):
                invoice_type = self.invoice.type
            else:
                invoice_type = getattr(self, 'invoice_type', None)

            if invoice_type == 'in':
                if context.get('_deductible_rate') is not None:
                    deductible_rate = context['_deductible_rate']
                else:
                    deductible_rate = getattr(self, 'taxes_deductible_rate', 1)
                if deductible_rate is None:
                    deductible_rate = 1
                if not deductible_rate:
                    return []
            else:
                deductible_rate = 1
            return [(
                getattr(self, 'taxes', None) or [],
                (getattr(
                    self,
                    'unit_gross_price',
                    None) or Decimal(0)) * deductible_rate,
                getattr(self, 'quantity', None) or 0,
                getattr(self, 'tax_date', None),
                )]
        else:
            raise ValueError('invalid "modegross"')

    @fields.depends(
        'unit_price', 'unit_gross_price', 'taxes',
        'invoice', '_parent_invoice.invoice_date')
    def on_change_unit_gross_price(self):
        """ compute unit_price from gross_price
        """
        pool = Pool()
        Tax = pool.get('account.tax')
        IrDate = pool.get('ir.date')

        if not isinstance(self.unit_gross_price, type(None)):
            invoice_date = getattr(self.invoice, 'invoice_date', None) or \
                IrDate.today()
            self.unit_price = Tax.reverse_compute(
                self.unit_gross_price,
                self.taxes,
                date=invoice_date).quantize(
                    Decimal(1) / 10 ** self.__class__.unit_price.digits[1])

    @fields.depends(
        'unit_price', 'unit_gross_price', 'taxes', 'invoice',
        '_parent_invoice.invoice_date',
        methods=['on_change_unit_gross_price'])
    def on_change_taxes(self):
        self.on_change_unit_gross_price()

    @fields.depends('invoice', '_parent_invoice.modegross')
    def on_change_with_modegross(self, name=None):
        """ get mode of invoice
        """
        return getattr(self.invoice, 'modegross', 'net')

    @fields.depends(
        'type', 'quantity', 'unit_price', 'taxes_deductible_rate', 'invoice',
        '_parent_invoice.currency', 'currency', 'taxes', 'unit_gross_price',
        '_parent_invoice.type', 'invoice_type', 'modegross',
        methods=['_get_taxes'])
    def on_change_with_amount(self):
        """ compute amount of line
        """
        if self.modegross == 'net':
            return super(InvoiceLine, self).on_change_with_amount()
        elif self.modegross == 'gross':
            if self.type != 'line':
                return Decimal('0.0')
            currency = self.invoice.currency if self.invoice else self.currency
            amount = Decimal(str(self.quantity or '0.0')) * \
                (self.unit_gross_price or Decimal('0.0'))
            if currency:
                return currency.round(amount)
            return amount
        else:
            return Decimal('0.0')

# end InvoiceLine


class InvoiceLineDiscount(metaclass=PoolMeta):
    __name__ = 'account.invoice.line'

    @fields.depends('product', 'unit', 'modegross')
    def compute_base_price(self):
        """ switch net/gross
        """
        pool = Pool()
        Uom = pool.get('product.uom')

        if self.modegross != 'gross':
            return super(InvoiceLineDiscount, self).compute_base_price()
        else:
            if self.product:
                price = self.product.template.used_gross_price
                if price is not None:
                    if self.unit is not None:
                        price = Uom.compute_price(
                            self.product.default_uom, price, self.unit)
                    return round_price(price)
                return price

    @fields.depends(
        'unit_price', 'unit_gross_price', 'base_price', 'modegross')
    def on_change_with_discount_rate(self, name=None):
        """ switch net/gross
        """
        if self.modegross != 'gross':
            return super(
                InvoiceLineDiscount,
                self).on_change_with_discount_rate()
        else:
            if self.unit_gross_price is None or not self.base_price:
                return
            rate = 1 - self.unit_gross_price / self.base_price
            return rate.quantize(
                Decimal(1) / 10 ** self.__class__.discount_rate.digits[1])

    @fields.depends(
        'base_price', 'discount_rate', 'modegross', 'unit_gross_price',
        'taxes', 'invoice', '_parent_invoice.invoice_date',
        methods=[
            'on_change_with_discount_amount', 'on_change_with_discount',
            'on_change_with_amount'])
    def on_change_discount_rate(self):
        """ switch net/gross
        """
        Tax = Pool().get('account.tax')

        if self.modegross != 'gross':
            super(InvoiceLineDiscount, self).on_change_discount_rate()
        else:
            if self.base_price is not None and self.discount_rate is not None:
                self.unit_gross_price = round_price(
                    self.base_price * (1 - self.discount_rate))
                self.discount_amount = self.on_change_with_discount_amount()
                self.discount = self.on_change_with_discount()
                self.amount = self.on_change_with_amount()
                self.unit_price = Tax.reverse_compute(
                    self.unit_gross_price, self.taxes,
                    date=self.invoice.invoice_date).quantize(
                        Decimal(1) / 10 ** self.__class__.unit_price.digits[1])

    @fields.depends(
        'unit_price', 'unit_gross_price', 'base_price', 'modegross')
    def on_change_with_discount_amount(self, name=None):
        """ switch net/gross
        """
        if self.modegross != 'gross':
            return super(
                InvoiceLineDiscount,
                self).on_change_with_discount_amount()
        else:
            if self.unit_gross_price is None or self.base_price is None:
                return
            return round_price(self.base_price - self.unit_gross_price)

    @fields.depends(
        'base_price', 'discount_amount', 'modegross', 'unit_gross_price',
        'taxes', 'invoice', '_parent_invoice.invoice_date',
        methods=[
            'on_change_with_discount_rate', 'on_change_with_discount',
            'on_change_with_amount'])
    def on_change_discount_amount(self):
        """ switch net/gross
        """
        Tax = Pool().get('account.tax')

        if self.modegross != 'gross':
            super(InvoiceLineDiscount, self).on_change_discount_amount()
        else:
            if (self.base_price is not None) and \
                    (self.discount_amount is not None):
                self.unit_gross_price = round_price(
                    self.base_price - self.discount_amount)
                self.discount_rate = self.on_change_with_discount_rate()
                self.discount = self.on_change_with_discount()
                self.amount = self.on_change_with_amount()
                self.unit_price = Tax.reverse_compute(
                    self.unit_gross_price, self.taxes,
                    date=self.invoice.invoice_date).quantize(
                        Decimal(1) / 10 ** self.__class__.unit_price.digits[1])

# end InvoiceLineDiscount
