# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice-module for Tryton from m-ds.de.
# The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.model import fields, Index
from trytond.pool import PoolMeta, Pool
from trytond.pyson import Eval, Bool, And
from trytond.modules.product import price_digits, round_price
from .invoice import sel_modegross
from decimal import Decimal
from .tax import GrossTaxableMixin


class Purchase(GrossTaxableMixin, metaclass=PoolMeta):
    __name__ = 'purchase.purchase'

    modegross = fields.Selection(
        string='Purchase type',
        selection=sel_modegross,
        required=True,
        help='Calculate the purchase gross or net.',
        states={'readonly': Bool(Eval('lines'))}, depends=['lines'])

    @classmethod
    def __setup__(cls):
        super(Purchase, cls).__setup__()
        t = cls.__table__()
        cls._sql_indexes.update({
                Index(t, (t.modegross, Index.Equality())),
            })

    @classmethod
    def default_modegross(cls):
        """ get setting from config
        """
        Configuration = Pool().get('account.configuration')

        cfg1 = Configuration.get_singleton()
        return getattr(cfg1, 'modegross', 'net')

    @classmethod
    def get_amount(cls, sales, names):
        """ fixup amounts for gross-purchases in 'draft'-state
        """
        Purchase2 = Pool().get('purchase.purchase')

        res1 = super(Purchase, cls).get_amount(
            sales,
            ['total_amount', 'untaxed_amount', 'tax_amount'])

        purchase_lst = Purchase2.search([
                ('id', 'in', list(res1['total_amount'].keys())),
                ('modegross', '=', 'gross'),
                ['OR',
                    ('state', 'not in', cls._states_cached),
                    ('untaxed_amount_cache', '=', None),
                    ('tax_amount_cache', '=', None),
                    ('total_amount_cache', '=', None)]])

        for purchase in purchase_lst:
            res1['total_amount'][purchase.id] = sum(
                    (line.amount for line in purchase.lines
                        if line.type == 'line'), Decimal(0))
            res1['untaxed_amount'][purchase.id] = \
                res1['total_amount'][purchase.id] + \
                res1['tax_amount'][purchase.id]

        for key in list(res1.keys()):
            if key not in names:
                del res1[key]
        return res1

    @fields.depends(
        'modegross', 'lines', 'currency', methods=['get_tax_amount'])
    def on_change_lines(self):
        """ compute amounts, switch net/gross
        """
        self.untaxed_amount = Decimal('0.0')
        self.tax_amount = Decimal('0.0')
        self.total_amount = Decimal('0.0')

        if self.modegross == 'net':
            super(Purchase, self).on_change_lines()
        elif self.modegross == 'gross':
            if self.lines:
                for line in self.lines:
                    self.total_amount += getattr(line, 'amount', None) or 0
                self.tax_amount = self.get_tax_amount()
            if self.currency:
                self.total_amount = self.currency.round(self.total_amount)
                self.tax_amount = self.currency.round(self.tax_amount)
            self.untaxed_amount = self.total_amount - self.tax_amount
            if self.currency:
                self.untaxed_amount = self.currency.round(self.untaxed_amount)

    @property
    def taxable_lines(self):
        """ swtich net/gross
        """
        if self.modegross == 'gross':
            taxable_lines = []
            for line in self.lines:
                if getattr(line, 'type', None) != 'line':
                    continue
                taxable_lines.append((
                        getattr(line, 'taxes', None) or [],
                        getattr(line, 'unit_gross_price', None) or Decimal(0),
                        getattr(line, 'quantity', None) or 0,
                        None))
            return taxable_lines
        else:
            return super(Purchase, self).taxable_lines

    def _get_invoice_purchase(self):
        """ add 'modegross' to invoice
        """
        invoice = super(Purchase, self)._get_invoice_purchase()
        invoice.modegross = self.modegross
        return invoice

# end Purchase


class PurchaseLine(GrossTaxableMixin, metaclass=PoolMeta):
    __name__ = 'purchase.line'

    modegross = fields.Function(
        fields.Char(
            string='Purchase type', readonly=True,
            states={'invisible': True}),
        'on_change_with_modegross')
    unit_gross_price = fields.Numeric(
        'Gross Price',
        digits=price_digits,
        states={
            'invisible': ~And(
                Eval('type') == 'line',
                Eval('modegross', '').in_(['-', 'gross'])),
            'required': And(
                Eval('type') == 'line',
                Eval('modegross', '') == 'gross'),
            'readonly': Eval('purchase_state') != 'draft'},
        depends=['type', 'modegross', 'purchase_state'])

    @classmethod
    def __setup__(cls):
        super(PurchaseLine, cls).__setup__()
        cls.unit_price.states.update({
            'required': And(
                Eval('type', '') == 'line',
                Eval('modegross', '') == 'net'),
            })
        cls.unit_price.depends.add('modegross')

    def get_amount(self, name):
        """ fix funktion to use 'on_change_with_amount' on
            every call
        """
        if self.type == 'line':
            return self.on_change_with_amount()
        elif self.type == 'subtotal':
            amount = Decimal('0.0')
            for line2 in self.sale.lines:
                if line2.type == 'line':
                    amount += line2.on_change_with_amount()
                elif line2.type == 'subtotal':
                    if self == line2:
                        break
                    amount = Decimal('0.0')
            return amount
        return Decimal('0.0')

    def get_invoice_line(self):
        """ add gross-price
        """
        lines = super(PurchaseLine, self).get_invoice_line()

        if self.modegross == 'gross':
            for line in lines:
                line.unit_gross_price = self.unit_gross_price
                line.on_change_unit_gross_price()
        return lines

    @fields.depends(
        'product', 'unit', 'purchase', 'modegross',
        '_parent_purchase.party', '_parent_purchase.invoice_party',
        'product_supplier',
        methods=[
            'compute_taxes', 'compute_unit_price',
            '_get_product_supplier_pattern'])
    def on_change_product(self):
        """ add gross-price
        """
        super(PurchaseLine, self).on_change_product()
        if not self.product:
            return
        if self.product.template.used_gross_price is not None:
            self.unit_gross_price = \
                round_price(self.product.template.used_gross_price)
        self.on_change_unit_gross_price()
        self.amount = self.on_change_with_amount()

    @fields.depends('purchase', '_parent_purchase.modegross')
    def on_change_with_modegross(self, name=None):
        """ get mode of purchase
        """
        if self.purchase:
            return self.purchase.modegross
        return 'net'

    @fields.depends(
        'unit_price', 'unit_gross_price', 'taxes',
        'purchase', '_parent_purchase.purchase_date')
    def on_change_unit_gross_price(self):
        """ compute unit_price from gross_price
        """
        pool = Pool()
        Tax = pool.get('account.tax')
        IrDate = pool.get('ir.date')

        if not isinstance(self.unit_gross_price, type(None)):
            purchase_date = getattr(self.purchase, 'purchase_date', None)
            self.unit_price = Tax.reverse_compute(
                self.unit_gross_price, self.taxes,
                date=purchase_date or IrDate.today()).quantize(
                    Decimal(1) / 10 ** self.__class__.unit_price.digits[1])

    @fields.depends(
        'unit_price', 'unit_gross_price', 'taxes', 'purchase',
        '_parent_purchase.purchase_date')
    def on_change_taxes(self):
        """ update gross/net price
        """
        self.on_change_unit_gross_price()

    @fields.depends(
        'type', 'quantity', 'unit_price', 'unit', 'purchase',
        '_parent_purchase.currency', 'currency', 'modegross',
        'unit_gross_price')
    def on_change_with_amount(self):
        """ compute amount of line
        """
        if self.modegross == 'net':
            return super(PurchaseLine, self).on_change_with_amount()
        elif self.modegross == 'gross':
            if self.type != 'line':
                return Decimal('0.0')

            currency = self.purchase.currency \
                if self.purchase else self.currency

            amount = (
                Decimal(str(self.quantity or '0.0')) *
                (self.unit_gross_price or Decimal('0.0')))
            if currency:
                return currency.round(amount)
            return amount
        else:
            return Decimal('0.0')

# end PurchaseLine
