# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice-module for Tryton from m-ds.de.
# The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.model import fields, Index
from trytond.pool import PoolMeta, Pool
from trytond.pyson import Eval, Bool, And
from trytond.modules.product import price_digits, round_price
from trytond.transaction import Transaction
from .invoice import sel_modegross
from decimal import Decimal
from .tax import GrossTaxableMixin


class Sale(GrossTaxableMixin, metaclass=PoolMeta):
    __name__ = 'sale.sale'

    modegross = fields.Selection(
        string='Sale type', selection=sel_modegross,
        required=True, help='Calculate the sale gross or net.',
        states={'readonly': Bool(Eval('lines'))}, depends=['lines'])

    @classmethod
    def __setup__(cls):
        super(Sale, cls).__setup__()
        t = cls.__table__()
        cls._sql_indexes.update({
                Index(t, (t.modegross, Index.Equality())),
            })

    @classmethod
    def default_modegross(cls):
        """ get setting from config
        """
        Configuration = Pool().get('account.configuration')

        cfg1 = Configuration.get_singleton()
        return getattr(cfg1, 'modegross', 'net')

    @classmethod
    def get_amount(cls, sales, names):
        """ fixup amounts for gross-sales in 'draft'-state
        """
        Sale2 = Pool().get('sale.sale')

        res1 = super(Sale, cls).get_amount(
            sales,
            ['total_amount', 'untaxed_amount', 'tax_amount'])

        sale_lst = Sale2.search([
                ('id', 'in', list(res1['total_amount'].keys())),
                ('modegross', '=', 'gross'),
                ['OR',
                    ('state', 'not in', cls._states_cached),
                    ('untaxed_amount_cache', '=', None),
                    ('tax_amount_cache', '=', None),
                    ('total_amount_cache', '=', None)],
            ])

        for sale in sale_lst:
            res1['total_amount'][sale.id] = sum(
                    (line.amount for line in sale.lines
                        if line.type == 'line'), Decimal(0))
            res1['untaxed_amount'][sale.id] = \
                res1['total_amount'][sale.id] - \
                res1['tax_amount'][sale.id]

        for key in list(res1.keys()):
            if key not in names:
                del res1[key]
        return res1

    @fields.depends(
        'modegross', 'lines', 'currency', methods=['get_tax_amount'])
    def on_change_lines(self):
        """ compute amounts, switch net/gross
        """
        self.untaxed_amount = Decimal('0.0')
        self.tax_amount = Decimal('0.0')
        self.total_amount = Decimal('0.0')

        if self.modegross == 'net':
            super(Sale, self).on_change_lines()
        elif self.modegross == 'gross':
            if self.lines:
                for line in self.lines:
                    self.total_amount += getattr(line, 'amount', None) or 0
                self.tax_amount = self.get_tax_amount()
            if self.currency:
                self.total_amount = self.currency.round(self.total_amount)
                self.tax_amount = self.currency.round(self.tax_amount)
            self.untaxed_amount = self.total_amount - self.tax_amount
            if self.currency:
                self.untaxed_amount = self.currency.round(self.untaxed_amount)

    def _get_invoice_sale(self):
        """ add 'modegross' to invoice
        """
        invoice = super(Sale, self)._get_invoice_sale()
        invoice.modegross = self.modegross
        return invoice

# end Sale


class SaleLine(GrossTaxableMixin, metaclass=PoolMeta):
    __name__ = 'sale.line'

    modegross = fields.Function(
        fields.Char(
            string='Sale type', readonly=True, states={'invisible': True}),
        'on_change_with_modegross')
    unit_gross_price = fields.Numeric(
        string='Gross Price',
        digits=price_digits,
        states={
            'invisible': ~And(
                Eval('type') == 'line',
                Eval('modegross', '').in_(['-', 'gross'])),
            'required': And(
                Eval('type') == 'line',
                Eval('modegross', '') == 'gross'),
            'readonly': Eval('sale_state') != 'draft'},
        depends=['type', 'modegross', 'sale_state'])

    @classmethod
    def __setup__(cls):
        super(SaleLine, cls).__setup__()
        cls.unit_price.states.update({
            'required': And(
                Eval('type', '') == 'line',
                Eval('modegross', '') == 'net'),
            })
        cls.unit_price.depends.add('modegross')

    @property
    def taxable_lines(self):
        """ get taxable line, switch gross/net calculation
        """
        if getattr(self, 'sale', None) and \
                getattr(self.sale, 'modegross', None):
            modegross = self.sale.modegross
        else:
            modegross = getattr(self, 'modegross', 'net')

        if modegross == 'net':
            return super(SaleLine, self).taxable_lines
        elif modegross == 'gross':
            if getattr(self, 'type', None) == 'line':
                return [(
                    getattr(self, 'taxes', None) or [],
                    getattr(self, 'unit_gross_price', None) or Decimal(0),
                    getattr(self, 'quantity', None) or 0,
                    None,
                    )]
            else:
                return []
        else:
            raise ValueError('invalid "modegross"')

    def get_amount(self, name):
        """ fix funktion to use 'on_change_with_amount' on
            every call
        """
        if self.type == 'line':
            return self.on_change_with_amount()
        elif self.type == 'subtotal':
            amount = Decimal('0.0')
            for line2 in self.sale.lines:
                if line2.type == 'line':
                    amount += line2.on_change_with_amount()
                elif line2.type == 'subtotal':
                    if self == line2:
                        break
                    amount = Decimal('0.0')
            return amount
        return Decimal('0.0')

    def get_invoice_line(self):
        """ add gross-price
        """
        lines = super(SaleLine, self).get_invoice_line()

        if self.modegross == 'gross':
            for line in lines:
                line.unit_gross_price = self.unit_gross_price
                line.on_change_unit_gross_price()
        return lines

    @fields.depends(
        'product', 'unit', 'sale', 'modegross',
        '_parent_sale.party', '_parent_sale.invoice_party',
        methods=[
            'compute_taxes', 'compute_unit_price', 'on_change_with_amount',
            '_get_context_sale_price'])
    def on_change_product(self):
        """ add gross-price
        """
        super(SaleLine, self).on_change_product()
        if not self.product:
            return

        if self.product.template.used_gross_price is not None:
            with Transaction().set_context(self._get_context_sale_price()):
                self.unit_gross_price = round_price(
                    self.product._get_sale_unit_gross_price())
        self.on_change_unit_gross_price()
        self.amount = self.on_change_with_amount()

    @fields.depends('sale', '_parent_sale.modegross')
    def on_change_with_modegross(self, name=None):
        """ get mode of sale
        """
        if self.sale:
            return self.sale.modegross
        return 'net'

    @fields.depends(
        'unit_price', 'unit_gross_price', 'taxes',
        'sale', '_parent_sale.sale_date')
    def on_change_unit_gross_price(self):
        """ compute unit_price from gross_price
        """
        pool = Pool()
        Tax = pool.get('account.tax')
        IrDate = pool.get('ir.date')

        if not isinstance(self.unit_gross_price, type(None)):
            sale_date = getattr(self.sale, 'sale_date', None) or IrDate.today()
            self.unit_price = Tax.reverse_compute(
                self.unit_gross_price, self.taxes,
                date=sale_date).quantize(
                    Decimal(1) / 10 ** self.__class__.unit_price.digits[1])

    @fields.depends(
        'unit_price', 'unit_gross_price', 'taxes', 'sale',
        '_parent_sale.sale_date', methods=['on_change_quantity'])
    def on_change_taxes(self):
        super(SaleLine, self).on_change_taxes()
        self.on_change_unit_gross_price()

    @fields.depends(
        'type', 'quantity', 'unit_price', 'unit', 'sale',
        '_parent_sale.currency', 'modegross', 'currency', 'unit_gross_price')
    def on_change_with_amount(self):
        """ compute amount of line
        """
        if self.modegross == 'net':
            return super(SaleLine, self).on_change_with_amount()
        elif self.modegross == 'gross':
            if self.type != 'line':
                return Decimal('0.0')
            currency = self.sale.currency if self.sale else self.currency
            amount = Decimal(str(self.quantity or '0.0')) * \
                (self.unit_gross_price or Decimal('0.0'))
            if currency:
                return currency.round(amount)
            return amount
        else:
            return Decimal('0.0')

# end SaleLine


class SaleLineDiscount(metaclass=PoolMeta):
    __name__ = 'sale.line'

    @fields.depends('product', 'unit', 'modegross')
    def compute_base_price(self):
        """ switch net/gross
        """
        pool = Pool()
        Uom = pool.get('product.uom')

        if self.modegross != 'gross':
            return super(SaleLineDiscount, self).compute_base_price()
        else:
            if self.product:
                price = self.product.template.used_gross_price
                if price is not None:
                    if self.unit:
                        price = Uom.compute_price(
                            self.product.default_uom, price, self.unit)
                    return round_price(price)
                return price

    @fields.depends(
        'unit_price', 'unit_gross_price', 'base_price', 'modegross')
    def on_change_with_discount_rate(self, name=None):
        """ switch net/gross
        """
        if self.modegross != 'gross':
            return super(SaleLineDiscount, self).on_change_with_discount_rate()
        else:
            if self.unit_gross_price is None or not self.base_price:
                return
            rate = 1 - self.unit_gross_price / self.base_price
            return rate.quantize(
                Decimal(1) / 10 ** self.__class__.discount_rate.digits[1])

    @fields.depends(
        'base_price', 'discount_rate', 'modegross', 'unit_gross_price',
        'taxes', 'sale', '_parent_sale.sale_date',
        methods=[
            'on_change_with_discount_amount', 'on_change_with_discount',
            'on_change_with_amount'])
    def on_change_discount_rate(self):
        """ switch net/gross
        """
        pool = Pool()
        Tax = pool.get('account.tax')
        IrDate = pool.get('ir.date')

        if self.modegross != 'gross':
            super(SaleLineDiscount, self).on_change_discount_rate()
        else:
            if self.base_price is not None and self.discount_rate is not None:
                self.unit_gross_price = round_price(
                    self.base_price * (1 - self.discount_rate))
                self.discount_amount = self.on_change_with_discount_amount()
                self.discount = self.on_change_with_discount()
                self.amount = self.on_change_with_amount()
                self.unit_price = Tax.reverse_compute(
                    self.unit_gross_price, self.taxes,
                    date=self.sale.sale_date or IrDate.today()).quantize(
                        Decimal(1) / 10 ** self.__class__.unit_price.digits[1])

    @fields.depends(
        'unit_price', 'unit_gross_price', 'base_price', 'modegross')
    def on_change_with_discount_amount(self, name=None):
        """ switch net/gross
        """
        if self.modegross != 'gross':
            return super(
                SaleLineDiscount,
                self).on_change_with_discount_amount()
        else:
            if self.unit_gross_price is None or self.base_price is None:
                return
            return round_price(self.base_price - self.unit_gross_price)

    @fields.depends(
        'base_price', 'discount_amount', 'modegross', 'unit_gross_price',
        'taxes', 'sale', '_parent_sale.sale_date',
        methods=[
            'on_change_with_discount_rate', 'on_change_with_discount',
            'on_change_with_amount'])
    def on_change_discount_amount(self):
        """ switch net/gross
        """
        pool = Pool()
        Tax = pool.get('account.tax')
        IrDate = pool.get('ir.date')

        if self.modegross != 'gross':
            super(SaleLineDiscount, self).on_change_discount_amount()
        else:
            if (self.base_price is not None) and \
                    (self.discount_amount is not None):
                self.unit_gross_price = round_price(
                    self.base_price - self.discount_amount)
                self.discount_rate = self.on_change_with_discount_rate()
                self.discount = self.on_change_with_discount()
                self.amount = self.on_change_with_amount()
                self.unit_price = Tax.reverse_compute(
                    self.unit_gross_price, self.taxes,
                    date=self.sale.sale_date or IrDate.today()).quantize(
                        Decimal(1) / 10 ** self.__class__.unit_price.digits[1])

# end SaleLineDiscount
