# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice-module for Tryton from m-ds.de.
# The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.model import fields
from trytond.pool import Pool
from trytond.transaction import Transaction
from collections import namedtuple
from decimal import Decimal

_GrossTaxableLine = namedtuple(
    '__GrossTaxableLine', (
        'taxes', 'unit_gross_price', 'quantity', 'tax_date'))


class GrossTaxableMixin(object):
    __slots__ = ()

    @fields.depends('modegross', methods=['_get_tax_context', '_round_taxes'])
    def _get_taxes(self):
        """ compute taxes in gross-mode
        """
        modegross = getattr(self, 'modegross', 'net')
        if modegross == 'net':
            return super()._get_taxes()
        elif modegross != 'gross':
            raise ValueError('invalid "modegross"')
        return self._get_taxes_gross()

    @fields.depends(methods=['_get_tax_context', '_round_taxes'])
    def _get_taxes_gross(self):
        """ compute taxes in gross-mode,
            tax-rounding: per line only
        """
        pool = Pool()
        Tax = pool.get('account.tax')
        IrDate = pool.get('ir.date')

        taxes = {}
        with Transaction().set_context(self._get_tax_context()):
            taxable_lines = [
                _GrossTaxableLine(*params) for params in self.taxable_lines]
            for line in taxable_lines:
                tax_date = line.tax_date or self.tax_date or IrDate.today()

                # compute net-price of line
                gross_total_of_line = line.unit_gross_price * \
                    Decimal(str(line.quantity or 0.0))
                net_total_of_line = Tax.reverse_compute(
                    gross_total_of_line,
                    Tax.browse(line.taxes),
                    tax_date)

                # use net-total-of-line to get list of taxes of line
                l_taxes = Tax.compute(
                    Tax.browse(line.taxes),
                    net_total_of_line, 1.0, tax_date)

                taxes_per_line = []
                for tax in l_taxes:
                    taxline = self._compute_tax_line(**tax)
                    if self.currency:
                        taxline['base'] = self.currency.round(taxline['base'])
                        taxline['amount'] = self.currency.round(
                            taxline['amount'])
                    taxes_per_line.append(taxline)

                # fixup rounding errors: base + sum(taxes) --> gross_price
                if len(taxes_per_line) > 0:
                    line_total = taxes_per_line[0]['base']
                    for taxline in taxes_per_line:
                        line_total += taxline['amount']

                    # add difference to last tax-line
                    taxes_per_line[-1]['amount'] = self.currency.round(
                            taxes_per_line[-1]['amount'] +
                            gross_total_of_line - line_total)

                for taxline in taxes_per_line:
                    if taxline not in taxes:
                        taxes[taxline] = taxline
                    else:
                        taxes[taxline]['base'] += taxline['base']
                        taxes[taxline]['amount'] += taxline['amount']
        return taxes

# end GrossTaxableMixin
