# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice module from m-ds.de for
# Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.tests.test_tryton import with_transaction
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.exceptions import UserError
from decimal import Decimal
from datetime import date


class InvoiceTestCase(object):
    'Test invoice module'

    @with_transaction()
    def test_spinv_create_invoice(self):
        """ create invoice, check config
        """
        pool = Pool()
        Invoice = pool.get('account.invoice')
        InvoiceLine = pool.get('account.invoice.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([('name', '=', 'Main Receivable')])
            self.assertEqual(len(acc_lst), 1)

            inv1 = Invoice(
                type='out',
                account=acc_lst[0],
                party=party1,
                lines=[
                    InvoiceLine(
                        type='comment',
                        description='text',
                        ),
                    ],
                )
            inv1.on_change_party()
            inv1.on_change_type()
            inv1.save()

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(inv_lst[0].modegross, 'net')
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(inv_lst[0].lines[0].modegross, 'net')

            inv_lst[0].modegross = 'gross'
            inv_lst[0].save()

            inv_lst2 = Invoice.search([])
            self.assertEqual(len(inv_lst2), 1)
            self.assertEqual(inv_lst2[0].modegross, 'gross')
            self.assertEqual(len(inv_lst2[0].lines), 1)
            self.assertEqual(inv_lst2[0].lines[0].modegross, 'gross')

    @with_transaction()
    def test_spinv_create_invoice_net_check_requireds(self):
        """ create invoice 'net', check 'required' of price-fields
        """
        pool = Pool()
        Invoice = pool.get('account.invoice')
        InvoiceLine = pool.get('account.invoice.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([
                    ('name', 'in', ['Main Receivable', 'Main Revenue'])
                ], order=[('name', 'ASC')])
            self.assertEqual(len(acc_lst), 2)
            self.assertEqual(acc_lst[0].name, 'Main Receivable')
            self.assertEqual(acc_lst[1].name, 'Main Revenue')

            inv1 = Invoice(
                type='out',
                modegross='net',
                account=acc_lst[0],
                party=party1,
                lines=[
                    InvoiceLine(
                        type='line',
                        description='text',
                        account=acc_lst[1],
                        quantity=1.0,
                        unit_price=Decimal('2.5'),
                        ),
                    ],
                )
            inv1.on_change_party()
            inv1.on_change_type()
            inv1.save()

            inv1.lines[0].unit_price = None
            self.assertRaisesRegex(
                UserError,
                r'A value is required for field "Unit Price" in ' +
                r'"Main Revenue @ \(\d+\)" of "Invoice Line".',
                inv1.lines[0].save)

    @with_transaction()
    def test_spinv_create_invoice_gross_check_requireds(self):
        """ create invoice 'gross', check 'required' of price-fields
        """
        pool = Pool()
        Invoice = pool.get('account.invoice')
        InvoiceLine = pool.get('account.invoice.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([
                    ('name', 'in', ['Main Receivable', 'Main Revenue'])
                ], order=[('name', 'ASC')])
            self.assertEqual(len(acc_lst), 2)
            self.assertEqual(acc_lst[0].name, 'Main Receivable')
            self.assertEqual(acc_lst[1].name, 'Main Revenue')

            inv1 = Invoice(
                type='out',
                modegross='gross',
                account=acc_lst[0],
                party=party1,
                lines=[
                    InvoiceLine(
                        type='line',
                        description='text',
                        account=acc_lst[1],
                        quantity=1.0,
                        unit_gross_price=Decimal('2.5'),
                        ),
                    ],
                )
            inv1.on_change_party()
            inv1.on_change_type()
            inv1.save()

            inv1.lines[0].unit_gross_price = None
            self.assertRaisesRegex(
                UserError,
                r'A value is required for field "Gross Price" in ' +
                r'"Main Revenue @ \(\d+\)" of "Invoice Line"',
                inv1.lines[0].save)

    @with_transaction()
    def test_spinv_create_invoice_gross_edit_gross_price(self):
        """ create invoice 'gross', edit unit_gross_price,
            check resulting unit_price
        """
        pool = Pool()
        Invoice = pool.get('account.invoice')
        InvoiceLine = pool.get('account.invoice.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('VSt 19%', Decimal('-0.19'))
            cat1.customer_taxes = [tax2]
            cat1.supplier_taxes = [tax3]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([
                    ('name', 'in', ['Main Receivable', 'Main Revenue'])
                ], order=[('name', 'ASC')])
            self.assertEqual(len(acc_lst), 2)
            self.assertEqual(acc_lst[0].name, 'Main Receivable')
            self.assertEqual(acc_lst[1].name, 'Main Revenue')

            inv1 = Invoice(
                type='out',
                modegross='gross',
                account=acc_lst[0],
                party=party1,
                invoice_date=date(2021, 10, 1),
                lines=[
                    InvoiceLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.0,
                        unit_gross_price=Decimal('2.5'),
                        ),
                    ],
                )
            inv1.on_change_party()
            inv1.on_change_type()
            inv1.lines[0].on_change_product()
            inv1.lines[0].on_change_unit_gross_price()
            inv1.save()

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(
                inv_lst[0].lines[0].unit_gross_price,
                Decimal('2.5'))
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('2.1'))
            self.assertEqual(inv_lst[0].lines[0].quantity, 1.0)

    @with_transaction()
    def test_spinv_create_invoice_net_amount_1x_tax(self):
        """ create invoice 'net', calculate amount + tax,
            add analytic-account to invoice-line
        """
        pool = Pool()
        Invoice = pool.get('account.invoice')
        InvoiceLine = pool.get('account.invoice.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')
        AnalyticEntry = pool.get('analytic.account.entry')
        Move = pool.get('account.move')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            (analytic_account,
                distri_account) = self.prep_spinv_analytic_account(company1)
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('VSt 19%', Decimal('-0.19'))
            cat1.customer_taxes = [tax2]
            cat1.supplier_taxes = [tax3]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([
                    ('name', 'in', ['Main Receivable', 'Main Revenue'])
                ], order=[('name', 'ASC')])
            self.assertEqual(len(acc_lst), 2)
            self.assertEqual(acc_lst[0].name, 'Main Receivable')
            self.assertEqual(acc_lst[1].name, 'Main Revenue')

            inv1 = Invoice(
                type='out',
                modegross='net',
                account=acc_lst[0],
                party=party1,
                invoice_date=date(2021, 10, 1),
                lines=[
                    InvoiceLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        unit_price=Decimal('16.8'),
                        analytic_accounts=[
                            AnalyticEntry(
                                root=analytic_account.root,
                                account=distri_account)],
                        ),
                    ],
                )
            inv1.on_change_party()
            inv1.on_change_type()
            inv1.lines[0].on_change_product()
            inv1.on_change_lines()
            inv1.save()

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('16.80'))
            self.assertEqual(inv_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(inv_lst[0].lines[0].amount, Decimal('25.20'))
            self.assertEqual(inv_lst[0].lines[0].taxable_lines, [
                ([tax2], Decimal('16.8'), 1.5, date(2021, 10, 1)),
                ])
            key19 = {
                'manual': False, 'description': 'USt 19%',
                'legal_notice': None, 'base': Decimal('25.20'),
                'amount': Decimal('4.79'), 'tax': tax2.id,
                'account': tax2.invoice_account.id}
            taxes = list(inv_lst[0]._get_taxes().keys())
            self.assertEqual(len(taxes), 1)
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')

            mv_lines = inv_lst[0].lines[0].get_move_lines()
            self.assertEqual(len(mv_lines), 1)
            self.assertEqual(mv_lines[0].origin.id, inv_lst[0].lines[0].id)
            self.assertEqual(mv_lines[0].credit, Decimal('25.2'))
            self.assertEqual(mv_lines[0].debit, 0)
            self.assertEqual(
                mv_lines[0].account.id,
                inv_lst[0].lines[0].account.id)

            self.assertEqual(inv_lst[0].untaxed_amount, Decimal('25.20'))
            self.assertEqual(inv_lst[0].tax_amount, Decimal('4.79'))
            self.assertEqual(inv_lst[0].total_amount, Decimal('29.99'))
            self.assertEqual(len(inv_lst[0].taxes), 1)
            self.assertEqual(inv_lst[0].taxes[0].description, 'USt 19%')
            self.assertEqual(inv_lst[0].taxes[0].account.name, 'Main Tax')
            self.assertEqual(inv_lst[0].taxes[0].base, Decimal('25.20'))
            self.assertEqual(inv_lst[0].taxes[0].amount, Decimal('4.79'))
            self.assertEqual(inv_lst[0].taxes[0].tax.name, 'USt 19%')

            # search in 'total_amount' - net-calculation
            self.assertEqual(Invoice.search_count([
                ('total_amount', '>', Decimal('29.80'))]), 1)
            self.assertEqual(Invoice.search_count([
                ('total_amount', '<', Decimal('30.00'))]), 1)
            self.assertEqual(Invoice.search_count([
                ('total_amount', '=', Decimal('29.99'))]), 1)
            self.assertEqual(Invoice.search_count([
                ('total_amount', '>', Decimal('30.00'))]), 0)
            self.assertEqual(Invoice.search_count([
                ('total_amount', '<', Decimal('29.00'))]), 0)

            # post invoice
            self.assertEqual(Move.search_count([]), 0)
            Invoice.validate(inv_lst)
            Invoice.post(inv_lst)

            move, = Move.search([])
            self.assertEqual(move.date, date(2021, 10, 1))
            self.assertEqual(len(move.lines), 3)

            self.assertEqual(move.lines[0].debit, Decimal('29.99'))
            self.assertEqual(move.lines[0].credit, Decimal('0.0'))
            self.assertEqual(move.lines[0].account.name, 'Main Receivable')
            self.assertEqual(len(move.lines[0].analytic_lines), 0)

            self.assertEqual(move.lines[1].debit, Decimal('0.0'))
            self.assertEqual(move.lines[1].credit, Decimal('4.79'))
            self.assertEqual(move.lines[1].account.name, 'Main Tax')
            self.assertEqual(len(move.lines[1].analytic_lines), 0)

            self.assertEqual(move.lines[2].debit, Decimal('0.0'))
            self.assertEqual(move.lines[2].credit, Decimal('25.2'))
            self.assertEqual(move.lines[2].account.name, 'Main Revenue')
            self.assertEqual(len(move.lines[2].analytic_lines), 2)

            # 15.12 + 10.08 = 25.2
            self.assertEqual(
                move.lines[2].analytic_lines[0].debit, Decimal('0'))
            self.assertEqual(
                move.lines[2].analytic_lines[0].credit, Decimal('15.12'))
            self.assertEqual(
                move.lines[2].analytic_lines[0].account.name, 'D2')

            self.assertEqual(
                move.lines[2].analytic_lines[1].debit, Decimal('0'))
            self.assertEqual(
                move.lines[2].analytic_lines[1].credit, Decimal('10.08'))
            self.assertEqual(
                move.lines[2].analytic_lines[1].account.name, 'D1')

    @with_transaction()
    def test_spinv_create_invoice_net_amount_2x_tax(self):
        """ create invoice 'net', calculate amount + tax
        """
        pool = Pool()
        Invoice = pool.get('account.invoice')
        InvoiceLine = pool.get('account.invoice.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('USt 7%', Decimal('0.07'))
            tax4 = self.prep_spinv_add_tax('VSt 19%', Decimal('-0.19'))
            tax5 = self.prep_spinv_add_tax('VSt 7%', Decimal('-0.07'))
            cat1.customer_taxes = [tax2, tax3]
            cat1.supplier_taxes = [tax4, tax5]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([
                    ('name', 'in', ['Main Receivable', 'Main Revenue'])
                ], order=[('name', 'ASC')])
            self.assertEqual(len(acc_lst), 2)
            self.assertEqual(acc_lst[0].name, 'Main Receivable')
            self.assertEqual(acc_lst[1].name, 'Main Revenue')

            inv1 = Invoice(
                type='out',
                modegross='net',
                account=acc_lst[0],
                party=party1,
                invoice_date=date(2021, 10, 1),
                lines=[
                    InvoiceLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        unit_price=Decimal('16.8'),
                        ),
                    ],
                )
            inv1.on_change_party()
            inv1.on_change_type()
            inv1.lines[0].on_change_product()
            inv1.on_change_lines()
            inv1.save()

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('16.80'))
            self.assertEqual(inv_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(inv_lst[0].lines[0].amount, Decimal('25.20'))
            self.assertEqual(
                inv_lst[0].lines[0].taxable_lines, [
                    (
                        [tax2, tax3],
                        Decimal('16.8'), 1.5,
                        date(2021, 10, 1)),
                ])
            key7 = {
                'manual': False, 'description': 'USt 7%',
                'legal_notice': None, 'base': Decimal('25.20'),
                'amount': Decimal('1.76'), 'tax': tax3.id,
                'account': tax3.invoice_account.id}
            key19 = {
                'manual': False, 'description': 'USt 19%',
                'legal_notice': None,
                'base': Decimal('25.20'), 'amount': Decimal('4.79'),
                'tax': tax2.id, 'account': tax2.invoice_account.id}
            taxes = list(inv_lst[0]._get_taxes().keys())
            self.assertEqual(len(taxes), 2)
            if not str(key7) in str(taxes):
                raise ValueError('"key7" not found')
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')

            self.assertEqual(inv_lst[0].untaxed_amount, Decimal('25.20'))
            self.assertEqual(inv_lst[0].tax_amount, Decimal('6.55'))
            self.assertEqual(inv_lst[0].total_amount, Decimal('31.75'))
            self.assertEqual(len(inv_lst[0].taxes), 2)
            self.assertEqual(inv_lst[0].taxes[0].description, 'USt 19%')
            self.assertEqual(inv_lst[0].taxes[0].account.name, 'Main Tax')
            self.assertEqual(inv_lst[0].taxes[0].base, Decimal('25.20'))
            self.assertEqual(inv_lst[0].taxes[0].amount, Decimal('4.79'))
            self.assertEqual(inv_lst[0].taxes[0].tax.name, 'USt 19%')
            self.assertEqual(inv_lst[0].taxes[1].description, 'USt 7%')
            self.assertEqual(inv_lst[0].taxes[1].account.name, 'Main Tax')
            self.assertEqual(inv_lst[0].taxes[1].base, Decimal('25.20'))
            self.assertEqual(inv_lst[0].taxes[1].amount, Decimal('1.76'))
            self.assertEqual(inv_lst[0].taxes[1].tax.name, 'USt 7%')

    @with_transaction()
    def test_spinv_create_invoice_gross_amount_1x_tax(self):
        """ create invoice 'gross', calculate amount + tax,
            add analytic-account to invoice-line
        """
        pool = Pool()
        Invoice = pool.get('account.invoice')
        InvoiceLine = pool.get('account.invoice.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')
        AnalyticEntry = pool.get('analytic.account.entry')
        Move = pool.get('account.move')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            (analytic_account,
                distri_account) = self.prep_spinv_analytic_account(company1)
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax4 = self.prep_spinv_add_tax('VSt 19%', Decimal('-0.19'))
            cat1.customer_taxes = [tax2]
            cat1.supplier_taxes = [tax4]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([
                    ('name', 'in', ['Main Receivable', 'Main Revenue'])
                ], order=[('name', 'ASC')])
            self.assertEqual(len(acc_lst), 2)
            self.assertEqual(acc_lst[0].name, 'Main Receivable')
            self.assertEqual(acc_lst[1].name, 'Main Revenue')

            inv1 = Invoice(
                type='out',
                modegross='gross',
                account=acc_lst[0],
                party=party1,
                invoice_date=date(2021, 10, 1),
                lines=[
                    InvoiceLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        unit_gross_price=Decimal('20.0'),
                        analytic_accounts=[
                            AnalyticEntry(
                                root=analytic_account.root,
                                account=distri_account)],
                        ),
                    ],
                )
            inv1.on_change_party()
            inv1.on_change_type()
            inv1.lines[0].on_change_product()
            inv1.lines[0].unit_gross_price = Decimal('20.0')
            inv1.lines[0].base_price = Decimal('20.0')
            inv1.lines[0].on_change_unit_gross_price()
            inv1.on_change_lines()
            inv1.save()

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(
                inv_lst[0].lines[0].unit_gross_price,
                Decimal('20.0'))
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('16.81'))
            self.assertEqual(inv_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(inv_lst[0].lines[0].amount, Decimal('30.0'))
            self.assertEqual(inv_lst[0].lines[0].taxable_lines, [
                ((tax2,), Decimal('20.0'), 1.5, date(2021, 10, 1)),
                ])
            key19 = {
                'manual': False, 'description': 'USt 19%',
                'legal_notice': None,
                'base': Decimal('25.21'), 'amount': Decimal('4.79'),
                'tax': tax2.id, 'account': tax2.invoice_account.id}
            taxes = list(inv_lst[0]._get_taxes_gross().keys())
            self.assertEqual(len(taxes), 1)
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')

            mv_lines = inv_lst[0].lines[0].get_move_lines()
            self.assertEqual(len(mv_lines), 1)
            self.assertEqual(mv_lines[0].origin.id, inv_lst[0].lines[0].id)
            self.assertEqual(mv_lines[0].credit, Decimal('25.21'))
            self.assertEqual(mv_lines[0].debit, 0)
            self.assertEqual(
                mv_lines[0].account.id,
                inv_lst[0].lines[0].account.id)

            self.assertEqual(inv_lst[0].untaxed_amount, Decimal('25.21'))
            self.assertEqual(inv_lst[0].tax_amount, Decimal('4.79'))
            self.assertEqual(inv_lst[0].total_amount, Decimal('30.0'))
            self.assertEqual(len(inv_lst[0].taxes), 1)
            self.assertEqual(inv_lst[0].taxes[0].description, 'USt 19%')
            self.assertEqual(inv_lst[0].taxes[0].account.name, 'Main Tax')
            self.assertEqual(inv_lst[0].taxes[0].base, Decimal('25.21'))
            self.assertEqual(inv_lst[0].taxes[0].amount, Decimal('4.79'))
            self.assertEqual(inv_lst[0].taxes[0].tax.name, 'USt 19%')

            # search in 'total_amount' - gross-calculation
            self.assertEqual(Invoice.search_count([
                ('total_amount', '>', Decimal('29.80'))]), 1)
            self.assertEqual(Invoice.search_count([
                ('total_amount', '<', Decimal('30.20'))]), 1)
            self.assertEqual(Invoice.search_count([
                ('total_amount', '=', Decimal('30.00'))]), 1)
            self.assertEqual(Invoice.search_count([
                ('total_amount', '>', Decimal('30.00'))]), 0)
            self.assertEqual(Invoice.search_count([
                ('total_amount', '<', Decimal('30.00'))]), 0)

            self.assertEqual(
                inv_lst[0].lines[0].base_price,
                Decimal('20.0'))
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('16.81'))
            self.assertEqual(
                inv_lst[0].lines[0].unit_gross_price,
                Decimal('20.0'))

            # apply discount 10%
            inv_lst[0].lines[0].discount_rate = Decimal('0.1')
            inv_lst[0].lines[0].on_change_discount_rate()
            inv_lst[0].lines[0].save()
            self.assertEqual(
                inv_lst[0].lines[0].base_price,
                Decimal('20.0'))
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('15.13'))
            self.assertEqual(
                inv_lst[0].lines[0].unit_gross_price,
                Decimal('18.0'))

            # post invoice
            self.assertEqual(Move.search_count([]), 0)
            Invoice.validate(inv_lst)
            Invoice.post(inv_lst)

            move, = Move.search([])
            self.assertEqual(move.date, date(2021, 10, 1))
            self.assertEqual(len(move.lines), 3)

            self.assertEqual(move.lines[0].debit, Decimal('27.0'))
            self.assertEqual(move.lines[0].credit, Decimal('0.0'))
            self.assertEqual(move.lines[0].account.name, 'Main Receivable')
            self.assertEqual(len(move.lines[0].analytic_lines), 0)

            self.assertEqual(move.lines[1].debit, Decimal('0.0'))
            self.assertEqual(move.lines[1].credit, Decimal('4.31'))
            self.assertEqual(move.lines[1].account.name, 'Main Tax')
            self.assertEqual(len(move.lines[1].analytic_lines), 0)

            self.assertEqual(move.lines[2].debit, Decimal('0.0'))
            self.assertEqual(move.lines[2].credit, Decimal('22.69'))
            self.assertEqual(move.lines[2].account.name, 'Main Revenue')
            self.assertEqual(len(move.lines[2].analytic_lines), 2)

            # 13.61 + 9.08 = 22.69
            self.assertEqual(
                move.lines[2].analytic_lines[0].debit, Decimal('0'))
            self.assertEqual(
                move.lines[2].analytic_lines[0].credit, Decimal('13.61'))
            self.assertEqual(
                move.lines[2].analytic_lines[0].account.name, 'D2')

            self.assertEqual(
                move.lines[2].analytic_lines[1].debit, Decimal('0'))
            self.assertEqual(
                move.lines[2].analytic_lines[1].credit, Decimal('9.08'))
            self.assertEqual(
                move.lines[2].analytic_lines[1].account.name, 'D1')

    @with_transaction()
    def test_spinv_create_invoice_gross_amount_2x_tax(self):
        """ create invoice 'gross', calculate amount + tax
        """
        pool = Pool()
        Invoice = pool.get('account.invoice')
        InvoiceLine = pool.get('account.invoice.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('USt 7%', Decimal('0.07'))
            tax4 = self.prep_spinv_add_tax('VSt 19%', Decimal('-0.19'))
            tax5 = self.prep_spinv_add_tax('VSt 7%', Decimal('-0.07'))
            cat1.customer_taxes = [tax2, tax3]
            cat1.supplier_taxes = [tax4, tax5]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([
                    ('name', 'in', ['Main Receivable', 'Main Revenue'])
                ], order=[('name', 'ASC')])
            self.assertEqual(len(acc_lst), 2)
            self.assertEqual(acc_lst[0].name, 'Main Receivable')
            self.assertEqual(acc_lst[1].name, 'Main Revenue')

            inv1 = Invoice(
                type='out',
                modegross='gross',
                account=acc_lst[0],
                party=party1,
                invoice_date=date(2021, 10, 1),
                lines=[
                    InvoiceLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        unit_gross_price=Decimal('20.0'),
                        ),
                    ],
                )
            inv1.on_change_party()
            inv1.on_change_type()
            inv1.lines[0].on_change_product()
            inv1.lines[0].on_change_unit_gross_price()
            inv1.on_change_lines()
            inv1.save()

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(
                inv_lst[0].lines[0].unit_gross_price,
                Decimal('20.0'))
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('15.87'))
            self.assertEqual(inv_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(inv_lst[0].lines[0].amount, Decimal('30.0'))
            self.assertEqual(inv_lst[0].lines[0].taxable_lines, [
                ((tax2, tax3,), Decimal('20.0'), 1.5, date(2021, 10, 1)),
                ])
            key7 = {
                'manual': False, 'description': 'USt 7%',
                'legal_notice': None,
                'base': Decimal('23.81'), 'amount': Decimal('1.67'),
                'tax': tax3.id, 'account': tax3.invoice_account.id}
            key19 = {
                'manual': False, 'description': 'USt 19%',
                'legal_notice': None,
                'base': Decimal('23.81'), 'amount': Decimal('4.52'),
                'tax': tax2.id, 'account': tax2.invoice_account.id}
            taxes = list(inv_lst[0]._get_taxes_gross().keys())
            self.assertEqual(len(taxes), 2)
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')
            if not str(key7) in str(taxes):
                raise ValueError('"key7" not found')

            self.assertEqual(inv_lst[0].untaxed_amount, Decimal('23.81'))
            self.assertEqual(inv_lst[0].tax_amount, Decimal('6.19'))
            self.assertEqual(inv_lst[0].total_amount, Decimal('30.0'))
            self.assertEqual(len(inv_lst[0].taxes), 2)
            self.assertEqual(inv_lst[0].taxes[0].description, 'USt 19%')
            self.assertEqual(inv_lst[0].taxes[0].account.name, 'Main Tax')
            self.assertEqual(inv_lst[0].taxes[0].base, Decimal('23.81'))
            self.assertEqual(inv_lst[0].taxes[0].amount, Decimal('4.52'))
            self.assertEqual(inv_lst[0].taxes[0].tax.name, 'USt 19%')
            self.assertEqual(inv_lst[0].taxes[1].description, 'USt 7%')
            self.assertEqual(inv_lst[0].taxes[1].account.name, 'Main Tax')
            self.assertEqual(inv_lst[0].taxes[1].base, Decimal('23.81'))
            self.assertEqual(inv_lst[0].taxes[1].amount, Decimal('1.67'))
            self.assertEqual(inv_lst[0].taxes[1].tax.name, 'USt 7%')

# end InvoiceTestCase
