# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice module from m-ds.de for
# Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.tests.test_tryton import with_transaction
from trytond.pool import Pool
from trytond.transaction import Transaction
from decimal import Decimal
from datetime import date


class PriceListTestCase(object):
    'Test price list module'

    @with_transaction()
    def test_spinv_check_get_context_formula(self):
        """ create pricelist, check insert of 'gross_price' into
            formula-context
        """
        pool = Pool()
        PriceList = pool.get('product.price_list')
        Product = pool.get('product.template')
        ProductUom = pool.get('product.uom')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            uom, = ProductUom.search([('symbol', '=', 'u')])
            product, = Product.create([{
                'name': 'Part 1',
                'gross_price': Decimal('5.0'),
                'list_price': Decimal('3.0'),
                'type': 'service',
                'default_uom': uom.id,
                'products': [('create', [{}])],
                }])
            pricelist, = PriceList.create([{
                'name': 'Pl1',
                'tax_included': False,
                'price': 'list_price',
                'lines': [('create', [{
                    'product': product.products[0].id,
                    'formula': 'gross_price + 10.0',
                    }])],
                }])

            # check gross_price in context
            result = pricelist.get_context_formula(
                product.products[0], None, None, None)
            self.assertEqual(result['names']['gross_price'], Decimal('5.0'))

            # compute
            result = pricelist.compute(product.products[0], None, uom)
            # gross_price + 10.0 --> 15.0
            self.assertEqual(result, Decimal('15.0'))

    @with_transaction()
    def test_spinv_sale_net_price_list(self):
        """ create sale 'net', add price-list, calculate amount + tax
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')
        PriceList = pool.get('product.price_list')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            cat1.customer_taxes = [tax2]
            cat1.supplier_taxes = [tax3]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)
            prod1.sale_uom = prod1.default_uom
            prod1.salable = True
            prod1.list_price = Decimal('100.0')
            prod1.gross_price = None
            prod1.save()
            self.assertEqual(prod1.list_price, Decimal('100.0'))
            self.assertEqual(prod1.gross_price, None)

            price_list, = PriceList.create([{
                'name': 'Discount 10%',
                'company': company1.id,
                'tax_included': False,
                'price': 'list_price',
                'lines': [('create', [{
                    'product': prod1.id,
                    'formula': 'unit_price - unit_price * 0.1',
                    }])],
                }])

            sale1, = Sale.create([{
                'modegross': 'net',
                'party': Party.create([{
                    'name': 'Customer',
                    'addresses': [('create', [{
                        'city': 'City',
                        'street': 'Street 1',
                        'postal_code': '12345',
                        }])],
                    }])[0].id,
                'sale_date': date(2021, 10, 1),
                'price_list': price_list.id,
                }])
            self.assertEqual(len(sale1.lines), 0)
            self.assertEqual(sale1.party.rec_name, 'Customer')
            self.assertEqual(sale1.price_list.rec_name, 'Discount 10%')
            self.assertEqual(
                sale1.price_list.lines[0].formula,
                'unit_price - unit_price * 0.1')
            self.assertEqual(sale1.price_list.lines[0].category, None)
            self.assertEqual(
                sale1.price_list.lines[0].product.rec_name,
                'Pen')

            sale1.lines = [
                SaleLine(
                    type='line',
                    product=prod1.products[0],
                    description='text',
                    quantity=1.0,
                    ),
                ]
            sale1.lines[0].on_change_product()
            sale1.on_change_lines()
            sale1.save()

            sale_lst = Sale.search([])
            self.assertEqual(len(sale_lst[0].lines), 1)
            self.assertEqual(
                sale_lst[0].lines[0].base_price,
                Decimal('100.00'))
            self.assertEqual(
                sale_lst[0].lines[0].unit_price,
                Decimal('90.00'))
            self.assertEqual(sale_lst[0].lines[0].quantity, 1.0)
            self.assertEqual(sale_lst[0].lines[0].amount, Decimal('90.0'))
            self.assertEqual(sale_lst[0].lines[0].taxable_lines, [
                ((tax2,), Decimal('90.0'), 1.0, None),
                ])

    @with_transaction()
    def test_spinv_sale_gross_price_list(self):
        """ create sale 'gross', add price-list, calculate amount + tax
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')
        PriceList = pool.get('product.price_list')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            cat1.customer_taxes = [tax2]
            cat1.supplier_taxes = [tax3]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)
            prod1.sale_uom = prod1.default_uom
            prod1.salable = True
            prod1.gross_price = Decimal('100.0')
            prod1.on_change_gross_price()
            prod1.save()
            self.assertEqual(prod1.gross_price, Decimal('100.0'))
            self.assertEqual(prod1.list_price, Decimal('84.03'))

            price_list, = PriceList.create([{
                'name': 'Discount 10%',
                'company': company1.id,
                'tax_included': False,
                'price': 'list_price',
                'lines': [('create', [{
                    'product': prod1.id,
                    'formula': 'gross_price - gross_price * 0.1',
                    }])],
                }])

            sale1, = Sale.create([{
                'modegross': 'gross',
                'party': Party.create([{
                            'name': 'Customer',
                            'addresses': [('create', [{
                                'city': 'City',
                                'street': 'Street 1',
                                'postal_code': '12345',
                                }])],
                            }])[0].id,
                'sale_date': date(2021, 10, 1),
                'price_list': price_list.id,
                }])
            self.assertEqual(len(sale1.lines), 0)
            self.assertEqual(sale1.party.rec_name, 'Customer')
            self.assertEqual(sale1.price_list.rec_name, 'Discount 10%')
            self.assertEqual(
                sale1.price_list.lines[0].formula,
                'gross_price - gross_price * 0.1')
            self.assertEqual(sale1.price_list.lines[0].category, None)
            self.assertEqual(
                sale1.price_list.lines[0].product.rec_name,
                'Pen')

            sale1.lines = [
                SaleLine(
                    type='line',
                    product=prod1.products[0],
                    description='text',
                    quantity=1.0,
                    ),
                ]
            sale1.lines[0].on_change_product()
            sale1.on_change_lines()
            sale1.save()

            sale_lst = Sale.search([])
            self.assertEqual(len(sale_lst[0].lines), 1)
            # gross-price
            self.assertEqual(
                sale_lst[0].lines[0].base_price,
                Decimal('100.00'))
            # discounted gross-price
            self.assertEqual(
                sale_lst[0].lines[0].unit_gross_price,
                Decimal('90.00'))
            # net, dicsounted without tax
            self.assertEqual(
                sale_lst[0].lines[0].unit_price,
                Decimal('75.63'))
            self.assertEqual(sale_lst[0].lines[0].quantity, 1.0)
            self.assertEqual(sale_lst[0].lines[0].amount, Decimal('90.0'))
            self.assertEqual(sale_lst[0].lines[0].taxable_lines, [
                ((tax2,), Decimal('90.0'), 1.0, None),
                ])

# PriceListTestCase
