# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice module from m-ds.de for
# Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.tests.test_tryton import with_transaction
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.exceptions import UserError
from decimal import Decimal
from datetime import date


class PurchaseTestCase(object):
    'Test purchase module'

    @with_transaction()
    def test_spinv_create_purchase(self):
        """ create purchase, check config
        """
        pool = Pool()
        Purchase = pool.get('purchase.purchase')
        PurchaseLine = pool.get('purchase.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([('name', '=', 'Main Receivable')])
            self.assertEqual(len(acc_lst), 1)

            pur1 = Purchase(
                party=party1,
                lines=[
                    PurchaseLine(
                        type='comment',
                        description='text',
                        ),
                    ],
                )
            pur1.on_change_party()
            pur1.on_change_lines()
            pur1.save()

            pur_lst = Purchase.search([])
            self.assertEqual(len(pur_lst), 1)
            self.assertEqual(pur_lst[0].modegross, 'net')
            self.assertEqual(len(pur_lst[0].lines), 1)
            self.assertEqual(pur_lst[0].lines[0].modegross, 'net')

            pur_lst[0].modegross = 'gross'
            pur_lst[0].save()

            pur_lst2 = Purchase.search([])
            self.assertEqual(len(pur_lst2), 1)
            self.assertEqual(pur_lst2[0].modegross, 'gross')
            self.assertEqual(len(pur_lst2[0].lines), 1)
            self.assertEqual(pur_lst2[0].lines[0].modegross, 'gross')

    @with_transaction()
    def test_spinv_create_purchase_gross_check_requireds(self):
        """ create purchase 'gross', check 'required' of price-fields
        """
        pool = Pool()
        Purchase = pool.get('purchase.purchase')
        PurchaseLine = pool.get('purchase.line')
        Party = pool.get('party.party')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            pur1 = Purchase(
                modegross='gross',
                party=party1,
                lines=[
                    PurchaseLine(
                        type='line',
                        description='text',
                        quantity=1.0,
                        unit_gross_price=Decimal('2.5'),
                        ),
                    ],
                )
            pur1.on_change_party()
            pur1.on_change_lines()
            pur1.save()

            pur1.lines[0].unit_gross_price = None
            self.assertRaisesRegex(
                UserError,
                r'A value is required for field "Gross Price" in ' +
                r'"\(2\)" of "Purchase Line".',
                pur1.lines[0].save)

    @with_transaction()
    def test_spinv_create_purchase_net_amount_1x_tax(self):
        """ create purchase 'net', calculate amount + tax
        """
        pool = Pool()
        Purchase = pool.get('purchase.purchase')
        PurchaseLine = pool.get('purchase.line')
        Party = pool.get('party.party')
        Invoice = pool.get('account.invoice')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            cat1.customer_taxes = [tax2]
            cat1.supplier_taxes = [tax3]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)
            prod1.purchase_uom = prod1.default_uom
            prod1.purchasable = True
            prod1.save()

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            pur1 = Purchase(
                modegross='net',
                party=party1,
                purchase_date=date(2021, 10, 1),
                lines=[
                    PurchaseLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        ),
                    ],
                )
            pur1.on_change_party()
            pur1.lines[0].on_change_product()
            pur1.lines[0].unit_price = Decimal('16.8')
            pur1.on_change_lines()
            pur1.save()

            pur_lst = Purchase.search([])
            self.assertEqual(len(pur_lst), 1)
            self.assertEqual(pur_lst[0].taxable_lines, [
                ((tax3,), Decimal('16.8'), 1.5, None),
                ])
            self.assertEqual(len(pur_lst[0].lines), 1)
            self.assertEqual(
                pur_lst[0].lines[0].unit_price,
                Decimal('16.80'))
            self.assertEqual(pur_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(pur_lst[0].lines[0].amount, Decimal('25.20'))
            key19 = {
                'manual': False, 'description': 'VSt 19%',
                'legal_notice': None,
                'base': Decimal('25.20'), 'amount': Decimal('4.79'),
                'tax': tax3.id, 'account': tax3.invoice_account.id}
            taxes = list(pur_lst[0]._get_taxes().keys())
            self.assertEqual(len(taxes), 1)
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')

            self.assertEqual(pur_lst[0].untaxed_amount, Decimal('25.20'))
            self.assertEqual(pur_lst[0].tax_amount, Decimal('4.79'))
            self.assertEqual(pur_lst[0].total_amount, Decimal('29.99'))

            self.assertEqual(Invoice.search_count([]), 0)

            Purchase.quote(pur_lst)
            Purchase.confirm(pur_lst)
            Purchase.process(pur_lst)

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(inv_lst[0].modegross, 'net')
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('16.8'))
            self.assertEqual(inv_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(inv_lst[0].lines[0].product.name, 'Pen')

    @with_transaction()
    def test_spinv_create_purchase_net_amount_2x_tax(self):
        """ create purchase 'net', calculate amount + tax
        """
        pool = Pool()
        Purchase = pool.get('purchase.purchase')
        PurchaseLine = pool.get('purchase.line')
        Party = pool.get('party.party')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('USt 7%', Decimal('0.07'))
            tax4 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            tax5 = self.prep_spinv_add_tax('VSt 7%', Decimal('0.07'))
            cat1.customer_taxes = [tax2, tax3]
            cat1.supplier_taxes = [tax4, tax5]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)
            prod1.purchase_uom = prod1.default_uom
            prod1.purchasable = True
            prod1.save()

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            pur1 = Purchase(
                modegross='net',
                party=party1,
                purchase_date=date(2021, 10, 1),
                lines=[
                    PurchaseLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        ),
                    ],
                )
            pur1.on_change_party()
            pur1.lines[0].on_change_product()
            pur1.lines[0].unit_price = Decimal('16.8')
            pur1.on_change_lines()
            pur1.save()

            pur_lst = Purchase.search([])
            self.assertEqual(len(pur_lst), 1)
            self.assertEqual(pur_lst[0].taxable_lines, [
                ((tax4, tax5, ), Decimal('16.8'), 1.5, None),
                ])
            self.assertEqual(len(pur_lst[0].lines), 1)
            self.assertEqual(
                pur_lst[0].lines[0].unit_price,
                Decimal('16.80'))
            self.assertEqual(pur_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(pur_lst[0].lines[0].amount, Decimal('25.20'))

            key7 = {
                'manual': False, 'description': 'VSt 7%',
                'legal_notice': None,
                'base': Decimal('25.20'), 'amount': Decimal('1.76'),
                'tax': tax5.id, 'account': tax5.invoice_account.id}
            key19 = {
                'manual': False, 'description': 'VSt 19%',
                'legal_notice': None,
                'base': Decimal('25.20'), 'amount': Decimal('4.79'),
                'tax': tax4.id, 'account': tax4.invoice_account.id}
            taxes = list(pur_lst[0]._get_taxes().keys())
            self.assertEqual(len(taxes), 2)
            if not str(key7) in str(taxes):
                raise ValueError('"key7" not found')
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')

            self.assertEqual(pur_lst[0].untaxed_amount, Decimal('25.20'))
            self.assertEqual(pur_lst[0].tax_amount, Decimal('6.55'))
            self.assertEqual(pur_lst[0].total_amount, Decimal('31.75'))

# end PurchaseTestCase
