# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice module from m-ds.de for
# Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.tests.test_tryton import with_transaction
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.exceptions import UserError
from decimal import Decimal
from datetime import date


class SaleTestCase(object):
    'Test sale module'

    @with_transaction()
    def test_spinv_create_sale(self):
        """ create sale, check config
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')
        Account = pool.get('account.account')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            acc_lst = Account.search([('name', '=', 'Main Receivable')])
            self.assertEqual(len(acc_lst), 1)

            sale1 = Sale(
                party=party1,
                lines=[
                    SaleLine(
                        type='comment',
                        description='text',
                        ),
                    ],
                )
            sale1.on_change_party()
            sale1.on_change_lines()
            sale1.save()

            sale_lst = Sale.search([])
            self.assertEqual(len(sale_lst), 1)
            self.assertEqual(sale_lst[0].modegross, 'net')
            self.assertEqual(len(sale_lst[0].lines), 1)
            self.assertEqual(sale_lst[0].lines[0].modegross, 'net')

            sale_lst[0].modegross = 'gross'
            sale_lst[0].save()

            sale_lst2 = Sale.search([])
            self.assertEqual(len(sale_lst2), 1)
            self.assertEqual(sale_lst2[0].modegross, 'gross')
            self.assertEqual(len(sale_lst2[0].lines), 1)
            self.assertEqual(sale_lst2[0].lines[0].modegross, 'gross')

    @with_transaction()
    def test_spinv_create_sale_net_check_requireds(self):
        """ create sale 'net', check 'required' of price-fields
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            sale1 = Sale(
                modegross='net',
                party=party1,
                lines=[
                    SaleLine(
                        type='line',
                        description='text',
                        quantity=1.0,
                        unit_price=Decimal('2.5'),
                        ),
                    ],
                )
            sale1.on_change_party()
            sale1.on_change_lines()
            sale1.save()

            sale1.lines[0].unit_price = None
            self.assertRaisesRegex(
                UserError,
                r'A value is required for field "Unit Price" in ' +
                r'"\(\d+\)" of "Sale Line',
                sale1.lines[0].save)

    @with_transaction()
    def test_spinv_create_sale_gross_check_requireds(self):
        """ create sale 'gross', check 'required' of price-fields
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            sale1 = Sale(
                modegross='gross',
                party=party1,
                lines=[
                    SaleLine(
                        type='line',
                        description='text',
                        quantity=1.0,
                        unit_gross_price=Decimal('2.5'),
                        ),
                    ],
                )
            sale1.on_change_party()
            sale1.on_change_lines()
            sale1.save()

            sale1.lines[0].unit_gross_price = None
            self.assertRaisesRegex(
                UserError,
                r'A value is required for field "Gross Price" in ' +
                r'"\(\d+\)" of "Sale Line".',
                sale1.lines[0].save)

    @with_transaction()
    def test_spinv_create_sale_net_amount_1x_tax(self):
        """ create sale 'net', calculate amount + tax
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')
        Invoice = pool.get('account.invoice')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            cat1.customer_taxes = [tax2]
            cat1.supplier_taxes = [tax3]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)
            prod1.sale_uom = prod1.default_uom
            prod1.salable = True
            prod1.save()

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            sale1 = Sale(
                modegross='net',
                party=party1,
                sale_date=date(2021, 10, 1),
                lines=[
                    SaleLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        ),
                    ],
                )
            sale1.on_change_party()
            sale1.lines[0].on_change_product()
            sale1.lines[0].unit_price = Decimal('16.8')
            sale1.on_change_lines()
            sale1.save()

            sale_lst = Sale.search([])
            self.assertEqual(len(sale_lst), 1)
            self.assertEqual(len(sale_lst[0].lines), 1)
            self.assertEqual(
                sale_lst[0].lines[0].unit_price,
                Decimal('16.80'))
            self.assertEqual(sale_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(sale_lst[0].lines[0].amount, Decimal('25.20'))
            self.assertEqual(sale_lst[0].lines[0].taxable_lines, [
                ((tax2,), Decimal('16.8'), 1.5, None),
                ])
            key19 = {
                'manual': False, 'description': 'USt 19%',
                'legal_notice': None,
                'base': Decimal('25.20'), 'amount': Decimal('4.79'),
                'tax': tax2.id, 'account': tax2.invoice_account.id}
            taxes = list(sale_lst[0]._get_taxes().keys())
            self.assertEqual(len(taxes), 1)
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')

            self.assertEqual(sale_lst[0].untaxed_amount, Decimal('25.20'))
            self.assertEqual(sale_lst[0].tax_amount, Decimal('4.79'))
            self.assertEqual(sale_lst[0].total_amount, Decimal('29.99'))

            self.assertEqual(Invoice.search_count([]), 0)

            Sale.quote(sale_lst)
            Sale.confirm(sale_lst)
            Sale.process(sale_lst)

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(inv_lst[0].modegross, 'net')
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('16.8'))
            self.assertEqual(inv_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(inv_lst[0].lines[0].product.name, 'Pen')

    @with_transaction()
    def test_spinv_create_sale_net_amount_2x_tax(self):
        """ create sale 'net', calculate amount + tax
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('USt 7%', Decimal('0.07'))
            tax4 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            tax5 = self.prep_spinv_add_tax('VSt 7%', Decimal('0.07'))
            cat1.customer_taxes = [tax2, tax3]
            cat1.supplier_taxes = [tax4, tax5]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)
            prod1.sale_uom = prod1.default_uom
            prod1.salable = True
            prod1.save()

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            sale1 = Sale(
                modegross='net',
                party=party1,
                sale_date=date(2021, 10, 1),
                lines=[
                    SaleLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        ),
                    ],
                )
            sale1.on_change_party()
            sale1.lines[0].on_change_product()
            sale1.lines[0].unit_price = Decimal('16.8')
            sale1.on_change_lines()
            sale1.save()

            sale_lst = Sale.search([])
            self.assertEqual(len(sale_lst), 1)
            self.assertEqual(len(sale_lst[0].lines), 1)
            self.assertEqual(
                sale_lst[0].lines[0].unit_price,
                Decimal('16.80'))
            self.assertEqual(sale_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(sale_lst[0].lines[0].amount, Decimal('25.20'))
            self.assertEqual(sale_lst[0].lines[0].taxable_lines, [
                ((tax2, tax3, ), Decimal('16.8'), 1.5, None),
                ])
            key7 = {
                'manual': False, 'description': 'USt 7%',
                'legal_notice': None, 'base': Decimal('25.20'),
                'amount': Decimal('1.76'), 'tax': tax3.id,
                'account': tax3.invoice_account.id}
            key19 = {
                'manual': False, 'description': 'USt 19%',
                'legal_notice': None, 'base': Decimal('25.20'),
                'amount': Decimal('4.79'), 'tax': tax2.id,
                'account': tax2.invoice_account.id}
            taxes = list(sale_lst[0]._get_taxes().keys())
            self.assertEqual(len(taxes), 2)
            if not str(key7) in str(taxes):
                raise ValueError('"key7" not found')
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')

            self.assertEqual(sale_lst[0].untaxed_amount, Decimal('25.20'))
            self.assertEqual(sale_lst[0].tax_amount, Decimal('6.55'))
            self.assertEqual(sale_lst[0].total_amount, Decimal('31.75'))

    @with_transaction()
    def test_spinv_create_sale_gross_amount_1x_tax(self):
        """ create sale 'gross', calculate amount + tax
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')
        Invoice = pool.get('account.invoice')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax4 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            cat1.customer_taxes = [tax2]
            cat1.supplier_taxes = [tax4]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)
            prod1.sale_uom = prod1.default_uom
            prod1.salable = True
            prod1.gross_price = Decimal('20.0')
            prod1.on_change_gross_price()
            prod1.save()

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            sale1 = Sale(
                modegross='gross',
                party=party1,
                sale_date=date(2021, 10, 1),
                lines=[
                    SaleLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        ),
                    ],
                )
            sale1.on_change_party()
            sale1.lines[0].on_change_product()
            sale1.on_change_lines()
            sale1.save()

            sale_lst = Sale.search([])
            self.assertEqual(len(sale_lst), 1)
            self.assertEqual(len(sale_lst[0].lines), 1)
            self.assertEqual(
                sale_lst[0].lines[0].base_price,
                Decimal('20.0'))
            self.assertEqual(
                sale_lst[0].lines[0].unit_gross_price,
                Decimal('20.0'))
            self.assertEqual(
                sale_lst[0].lines[0].unit_price,
                Decimal('16.81'))
            self.assertEqual(sale_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(sale_lst[0].lines[0].amount, Decimal('30.0'))
            self.assertEqual(sale_lst[0].lines[0].taxable_lines, [
                ((tax2,), Decimal('20.0'), 1.5, None),
                ])
            key19 = {
                'manual': False, 'description': 'USt 19%',
                'legal_notice': None, 'base': Decimal('25.21'),
                'amount': Decimal('4.79'), 'tax': tax2.id,
                'account': tax2.invoice_account.id}
            taxes = list(sale_lst[0]._get_taxes_gross().keys())
            self.assertEqual(len(taxes), 1)
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')

            self.assertEqual(sale_lst[0].untaxed_amount, Decimal('25.21'))
            self.assertEqual(sale_lst[0].tax_amount, Decimal('4.79'))
            self.assertEqual(sale_lst[0].total_amount, Decimal('30.0'))

            sale_lst[0].lines[0].discount_rate = Decimal('0.1')
            sale_lst[0].lines[0].on_change_discount_rate()
            sale_lst[0].lines[0].save()
            self.assertEqual(
                sale_lst[0].lines[0].base_price,
                Decimal('20.0'))
            self.assertEqual(
                sale_lst[0].lines[0].unit_price,
                Decimal('15.13'))
            self.assertEqual(
                sale_lst[0].lines[0].unit_gross_price,
                Decimal('18.0'))
            sale_lst[0].lines[0].discount_rate = Decimal('0.0')
            sale_lst[0].lines[0].on_change_discount_rate()
            sale_lst[0].lines[0].save()

            self.assertEqual(Invoice.search_count([]), 0)

            Sale.quote(sale_lst)
            Sale.confirm(sale_lst)
            Sale.process(sale_lst)

            inv_lst = Invoice.search([])
            self.assertEqual(len(inv_lst), 1)
            self.assertEqual(inv_lst[0].modegross, 'gross')
            self.assertEqual(len(inv_lst[0].lines), 1)
            self.assertEqual(
                inv_lst[0].lines[0].unit_price,
                Decimal('16.81'))
            self.assertEqual(
                inv_lst[0].lines[0].unit_gross_price,
                Decimal('20.0'))
            self.assertEqual(inv_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(inv_lst[0].lines[0].product.name, 'Pen')

    @with_transaction()
    def test_spinv_create_sale_gross_amount_2x_tax(self):
        """ create sale 'gross', calculate amount + tax
        """
        pool = Pool()
        Sale = pool.get('sale.sale')
        SaleLine = pool.get('sale.line')
        Party = pool.get('party.party')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):
            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('USt 7%', Decimal('0.07'))
            tax4 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            tax5 = self.prep_spinv_add_tax('VSt 7%', Decimal('0.07'))
            cat1.customer_taxes = [tax2, tax3]
            cat1.supplier_taxes = [tax4, tax5]
            cat1.save()
            prod1 = self.prep_spinv_add_product('Pen', cat1)
            prod1.sale_uom = prod1.default_uom
            prod1.salable = True
            prod1.gross_price = Decimal('20.0')
            prod1.on_change_gross_price()
            prod1.save()

            party1, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{
                    'city': 'City',
                    'street': 'Street 1',
                    'postal_code': '12345',
                    }])],
                }])

            sale1 = Sale(
                modegross='gross',
                party=party1,
                sale_date=date(2021, 10, 1),
                lines=[
                    SaleLine(
                        type='line',
                        product=prod1.products[0],
                        description='text',
                        quantity=1.5,
                        ),
                    ],
                )
            sale1.on_change_party()
            sale1.lines[0].on_change_product()
            sale1.on_change_lines()
            sale1.save()

            sale_lst = Sale.search([])
            self.assertEqual(len(sale_lst), 1)
            self.assertEqual(len(sale_lst[0].lines), 1)
            self.assertEqual(
                sale_lst[0].lines[0].unit_gross_price,
                Decimal('20.0'))
            self.assertEqual(
                sale_lst[0].lines[0].unit_price,
                Decimal('15.87'))
            self.assertEqual(sale_lst[0].lines[0].quantity, 1.5)
            self.assertEqual(sale_lst[0].lines[0].amount, Decimal('30.0'))
            self.assertEqual(sale_lst[0].lines[0].taxable_lines, [
                ((tax2, tax3,), Decimal('20.0'), 1.5, None),
                ])
            key7 = {
                'manual': False, 'description': 'USt 7%',
                'legal_notice': None, 'base': Decimal('23.81'),
                'amount': Decimal('1.67'), 'tax': tax3.id,
                'account': tax3.invoice_account.id}
            key19 = {
                'manual': False, 'description': 'USt 19%',
                'legal_notice': None,
                'base': Decimal('23.81'), 'amount': Decimal('4.52'),
                'tax': tax2.id, 'account': tax2.invoice_account.id}
            taxes = list(sale_lst[0]._get_taxes_gross().keys())
            self.assertEqual(len(taxes), 2)
            if not str(key19) in str(taxes):
                raise ValueError('"key19" not found')
            if not str(key7) in str(taxes):
                raise ValueError('"key7" not found')

            self.assertEqual(sale_lst[0].untaxed_amount, Decimal('23.81'))
            self.assertEqual(sale_lst[0].tax_amount, Decimal('6.19'))
            self.assertEqual(sale_lst[0].total_amount, Decimal('30.0'))

# end SaleTestCase
