# -*- coding: utf-8 -*-
# This file is part of the sale-point-invoice module from m-ds.de for
# Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.tests.test_tryton import with_transaction
from trytond.pool import Pool
from trytond.transaction import Transaction
from decimal import Decimal
from datetime import date


class SubscriptionTestCase(object):
    'Test subscription module'

    @with_transaction()
    def test_spinv_create_subscription_add_net_invoice(self):
        """ add subsription, add net-invoice, check values
        """
        pool = Pool()
        Subscription = pool.get('sale.subscription')
        SubscriptionLine = pool.get('sale.subscription.line')
        SubscriptionService = pool.get('sale.subscription.service')
        Consumption = pool.get('sale.subscription.line.consumption')
        Party = pool.get('party.party')
        RecurrenceRule = pool.get('sale.subscription.recurrence.rule.set')
        UOM = pool.get('product.uom')
        Product = pool.get('product.template')
        ProductCategory = pool.get('product.category')
        Invoice = pool.get('account.invoice')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):

            customer, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{}])]}])

            recurrence, = RecurrenceRule.create([{
                'name': 'monthly service',
                'rules': [('create', [{
                    'freq': 'monthly',
                    'interval': 1,
                    }])],
                }])
            unit, = UOM.search([('symbol', '=', 'u')])

            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            ProductCategory.write(*[
                [cat1],
                {
                    'customer_taxes': [('add', [tax2.id])],
                    'supplier_taxes': [('add', [tax3.id])],
                }])
            product = self.prep_spinv_add_product('Rent a bike', cat1)
            Product.write(*[
                [product],
                {'type': 'service'},
                ])
            self.assertEqual(product.list_price, Decimal('8.4'))
            self.assertEqual(product.gross_price, Decimal('10.0'))

            subscription_service, = SubscriptionService.create([{
                'product': product.id,
                'consumption_recurrence': recurrence.id,
                }])

            subscr, = Subscription.create([{
                'description': 'monthly service',
                'party': customer.id,
                'invoice_address': customer.addresses[0].id,
                'start_date': date(2022, 5, 5),
                'invoice_recurrence': recurrence.id,
                'lines': [('create', [{
                    'service': subscription_service.id,
                    'description': 'monthly rent of bicycle',
                    'quantity': 1.0,
                    'unit': unit.id,
                    'unit_price': Decimal('15.0'),
                    'consumption_recurrence': recurrence.id,
                    'start_date': date(2022, 5, 5),
                    }])],
                }])
            self.assertEqual(subscr.modegross, 'net')
            self.assertEqual(subscr.party.rec_name, 'Customer')
            self.assertEqual(
                subscr.invoice_recurrence.rec_name, 'monthly service')
            self.assertEqual(len(subscr.lines), 1)
            self.assertEqual(
                subscr.lines[0].rec_name[:-3], '1.00\xa0u Rent a bike @ ')

            # activate subscription
            Subscription.quote([subscr])
            Subscription.run([subscr])

            # create consumption lines
            self.assertEqual(Consumption.search_count([]), 0)
            SubscriptionLine.generate_consumption(date=date(2022, 7, 20))
            self.assertEqual(Consumption.search_count([]), 3)

            # create invoice from consumptions
            self.assertEqual(Invoice.search_count([]), 0)
            Subscription.generate_invoice(date=date(2022, 7, 20))
            invoice, = Invoice.search([])

            # check invoice
            self.assertEqual(invoice.modegross, 'net')
            self.assertEqual(invoice.invoice_date, date(2022, 5, 5))
            self.assertEqual(invoice.untaxed_amount, Decimal('45.0'))
            self.assertEqual(invoice.tax_amount, Decimal('8.55'))
            self.assertEqual(invoice.total_amount, Decimal('53.55'))

            self.assertEqual(len(invoice.lines), 1)
            self.assertEqual(invoice.lines[0].product.rec_name, 'Rent a bike')
            self.assertEqual(invoice.lines[0].unit_price, Decimal('15.0'))
            self.assertEqual(invoice.lines[0].unit_gross_price, None)
            self.assertEqual(invoice.lines[0].type, 'line')
            self.assertEqual(invoice.lines[0].quantity, 3.0)
            self.assertEqual(invoice.lines[0].amount, Decimal('45.0'))

    @with_transaction()
    def test_spinv_create_subscription_add_gross_invoice(self):
        """ add subsription, add gross-invoice, check values
        """
        pool = Pool()
        Subscription = pool.get('sale.subscription')
        SubscriptionLine = pool.get('sale.subscription.line')
        SubscriptionService = pool.get('sale.subscription.service')
        Consumption = pool.get('sale.subscription.line.consumption')
        Party = pool.get('party.party')
        RecurrenceRule = pool.get('sale.subscription.recurrence.rule.set')
        UOM = pool.get('product.uom')
        Product = pool.get('product.template')
        ProductCategory = pool.get('product.category')
        Invoice = pool.get('account.invoice')

        company1 = self.prep_spinv_configuration()
        with Transaction().set_context({'company': company1.id}):

            customer, = Party.create([{
                'name': 'Customer',
                'addresses': [('create', [{}])]}])

            recurrence, = RecurrenceRule.create([{
                'name': 'monthly service',
                'rules': [('create', [{
                    'freq': 'monthly',
                    'interval': 1,
                    }])],
                }])
            unit, = UOM.search([('symbol', '=', 'u')])

            cat1 = self.prep_spinv_create_prodcategory('ProdCat')
            tax2 = self.prep_spinv_add_tax('USt 19%', Decimal('0.19'))
            tax3 = self.prep_spinv_add_tax('VSt 19%', Decimal('0.19'))
            ProductCategory.write(*[
                [cat1],
                {
                    'customer_taxes': [('add', [tax2.id])],
                    'supplier_taxes': [('add', [tax3.id])],
                }])
            product = self.prep_spinv_add_product('Rent a bike', cat1)
            Product.write(*[
                [product],
                {'type': 'service'},
                ])
            self.assertEqual(product.list_price, Decimal('8.4'))
            self.assertEqual(product.gross_price, Decimal('10.0'))

            subscription_service, = SubscriptionService.create([{
                'product': product.id,
                'consumption_recurrence': recurrence.id,
                }])

            subscr, = Subscription.create([{
                'modegross': 'gross',
                'description': 'monthly service - gross',
                'party': customer.id,
                'invoice_address': customer.addresses[0].id,
                'start_date': date(2022, 5, 5),
                'invoice_recurrence': recurrence.id,
                'lines': [('create', [{
                    'service': subscription_service.id,
                    'description': 'monthly rent of bicycle',
                    'quantity': 1.0,
                    'unit': unit.id,
                    'unit_price': Decimal('15.0'),  # used as gross
                    'consumption_recurrence': recurrence.id,
                    'start_date': date(2022, 5, 5),
                    }])],
                }])
            self.assertEqual(subscr.modegross, 'gross')
            self.assertEqual(subscr.party.rec_name, 'Customer')
            self.assertEqual(
                subscr.invoice_recurrence.rec_name, 'monthly service')
            self.assertEqual(len(subscr.lines), 1)
            self.assertEqual(
                subscr.lines[0].rec_name[:-3], '1.00\xa0u Rent a bike @ ')

            # activate subscription
            Subscription.quote([subscr])
            Subscription.run([subscr])

            # create consumption lines
            self.assertEqual(Consumption.search_count([]), 0)
            SubscriptionLine.generate_consumption(date=date(2022, 7, 20))
            self.assertEqual(Consumption.search_count([]), 3)

            # create invoice from consumptions
            self.assertEqual(Invoice.search_count([]), 0)
            Subscription.generate_invoice(date=date(2022, 7, 20))
            invoice, = Invoice.search([])

            # check invoice
            self.assertEqual(invoice.modegross, 'gross')
            self.assertEqual(invoice.invoice_date, date(2022, 5, 5))
            self.assertEqual(invoice.untaxed_amount, Decimal('37.82'))
            self.assertEqual(invoice.tax_amount, Decimal('7.18'))
            self.assertEqual(invoice.total_amount, Decimal('45.0'))

            self.assertEqual(len(invoice.lines), 1)
            self.assertEqual(invoice.lines[0].product.rec_name, 'Rent a bike')
            self.assertEqual(invoice.lines[0].unit_price, Decimal('12.61'))
            self.assertEqual(
                invoice.lines[0].unit_gross_price, Decimal('15.0'))
            self.assertEqual(invoice.lines[0].type, 'line')
            self.assertEqual(invoice.lines[0].quantity, 3.0)
            self.assertEqual(invoice.lines[0].amount, Decimal('45.0'))

# end SubscriptionTestCase
